# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd


class ExecutionException(Exception):
    """Base class for exceptions for execution of programs in QAT."""

    ...
